import base64
import copy
from typing import Any
from typing import Dict
from typing import List


async def search_raw_launch_template_version(
    hub, ctx, name=None, resource_id: str = None
) -> Dict:
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.ec2.describe_launch_template_versions(
        ctx,
        LaunchTemplateId=resource_id if resource_id else None,
        LaunchTemplateName=name if resource_id is None else None,
        Versions=["$Latest"],
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def search_raw(
    hub, ctx, name, filters: List = None, resource_id: str = None
) -> Dict:
    """Fetch one or more launch templates from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(str):
            The name of the Idem state.
        resource_id(str, Optional):
            AWS launch template id to identify the resource.
        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_launch_templates

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.ec2.describe_launch_templates(
        ctx,
        Filters=boto3_filter,
        LaunchTemplateIds=[resource_id] if resource_id else None,
    )

    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update_launch_template_version(
    hub,
    ctx,
    launch_template_id: str,
    old_launch_template_data: Dict[str, Any],
    new_launch_template_data: Dict[str, Any],
):
    """Update launch template version"""
    result = dict(comment=(), result=True, ret={})
    if hub.tool.aws.state_comparison_utils.compare_dicts(
        source_dict=old_launch_template_data, target_dict=new_launch_template_data
    ):
        return result
    if ctx.get("test", False):
        result["ret"] = new_launch_template_data
        return result

    ret = await hub.exec.boto3.client.ec2.create_launch_template_version(
        ctx=ctx,
        LaunchTemplateId=launch_template_id,
        LaunchTemplateData=pre_process_launch_template_data(
            hub, chunk=new_launch_template_data
        ),
    )
    result["result"] = ret["result"]
    result["comment"] = (f"Update launch_template_version",)
    result["ret"] = ret["ret"]["LaunchTemplateVersion"]
    await hub.exec.boto3.client.ec2.modify_launch_template(
        ctx=ctx,
        DefaultVersion=str(result["ret"]["VersionNumber"]),
        LaunchTemplateId=launch_template_id,
    )
    return result


def pre_process_launch_template_data(hub, chunk):
    """Base64 encode launch template user data"""
    desired_state = copy.deepcopy(chunk)
    if desired_state.get("UserData"):
        user_data = desired_state["UserData"]
        desired_state["UserData"] = base64.b64encode(user_data.encode()).decode()

    return desired_state



from typing import Any
from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, Any]],
    new_value: List[Dict[str, Any]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in its "present" state

    - network_interfaces:
        - AssociatePublicIpAddress: true
          DeleteOnTermination: true
          Description: ''
          DeviceIndex: 0
          Groups:
          - sg-XXXXXXXX
          InterfaceType: interface
          Ipv6Addresses: []
          NetworkCardIndex: 0
          PrivateIpAddresses:
          - Primary: true
            PrivateIpAddress: 888.88.88.888
          SubnetId: subnet-XXXXXXXX

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    comments.append(f"Unable to modify network interfaces at this time")
    return True
    result = True
    new_interfaces = {interface["DeviceIndex"]: interface for interface in new_value}
    old_interfaces = {interface["DeviceIndex"]: interface for interface in old_value}

    # create new interfaces
    interfaces_to_create = set(new_interfaces.keys()) - set(old_interfaces.keys())
    for device_index in interfaces_to_create:
        result &= await hub.tool.aws.ec2.instance.update.network_interfaces.create(
            ctx,
            resource,
            comments,
            device_index=device_index,
            new_interfaces=new_interfaces,
        )

    # Delete interfaces that were removed from the state
    interfaces_to_delete = set(old_interfaces.keys()) - set(new_interfaces.keys())
    for device_index in interfaces_to_delete:
        result &= await hub.tool.aws.ec2.instance.update.network_interfaces.delete(
            ctx,
            resource,
            comments,
            device_index=device_index,
            old_interfaces=old_interfaces,
        )

    # Modify existing network interfaces as needed
    interfaces_to_modify = set(old_interfaces.keys()).intersection(
        new_interfaces.keys()
    )
    for device_index in interfaces_to_modify:
        result &= await hub.tool.aws.ec2.instance.update.network_interfaces.modify(
            ctx,
            resource,
            comments,
            device_index=device_index,
            old_value=old_value,
            new_value=new_value,
        )

    return result


async def create(
    hub,
    ctx,
    resource,
    comments: List[str],
    device_index: int,
    old_interfaces: List[Dict[str, Any]],
) -> bool:
    attributes = old_interfaces[device_index]

    # Create the new interface
    ret = await hub.exec.boto3.client.ec2.create_network_interface(
        ctx,
        Description=attributes.get("Description"),
        DeviceIndex=device_index,
        Groups=attributes.get("Groups"),
        InterfaceType=attributes.get("InterfaceType"),
        Ipv6Addresses=attributes.get("Ipv6Addresses"),
        PrivateIpAddresses=attributes.get("PrivateIpAddresses"),
        # Use the subnet defined in the network interface, but fallback to the one used by the instance
        SubnetId=attributes.get("SubnetId", resource.subnet_id),
    )
    if ret.comment:
        comments.append(ret.comment)
    if not ret.result:
        return False

    # Attach the new interface to the instance
    interface_id = ret.ret["NetworkInterface"]["NetworkInterfaceId"]
    ret = await hub.exec.boto3.client.ec2.attach_network_interface(
        ctx,
        DeviceIndex=device_index,
        InstanceId=resource.id,
        NetworkInterfaceId=interface_id,
        NetworkCardIndex=attributes.get("NetworkCardIndex"),
    )
    if ret.comment:
        comments.append(ret.comment)
    if not ret.result:
        return False


async def delete(
    hub,
    ctx,
    resource,
    comments: List[str],
    device_index: int,
    new_interfaces: List[Dict[str, Any]],
) -> bool:
    ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
        ctx,
        Filters=[
            {"Name": "attachment.instance-id", "Values": [resource.id]},
            {"Name": "attachment.device-index", "Values": [device_index]},
        ],
    )

    if not ret:
        comments.append(f"Interface is already absent from index: {device_index}")
        return True

    # There will only be one result
    attachment_id = next(iter(ret.ret["NetworkInterfaces"]))["Attachment"][
        "AttachmentId"
    ]

    ret = await hub.exec.boto3.client.ec2.detach_network_interface(
        ctx, AttachmentId=attachment_id, Force=False
    )
    if ret.comment:
        comments.append(ret.comment)
    if not ret:
        return False


async def modify(
    hub,
    ctx,
    resource,
    comments: List[str],
    device_index: int,
    old_value: List[Dict[str, Any]],
    new_value: List[Dict[str, Any]],
) -> bool:
    if old_value[device_index] == new_value[device_index]:
        return True

    ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
        ctx,
        Filters=[
            {"Name": "attachment.instance-id", "Values": [resource.id]},
            {"Name": "attachment.device-index", "Values": [str(device_index)]},
        ],
    )
    if ret.comment:
        comments.append(ret.comment)
    if not ret.result:
        return False

    # There will only be one result
    interface = next(iter(ret.ret["NetworkInterfaces"]))

    # Check for changes to groups
    if old_value[device_index].get("Groups") != new_value[device_index].get("Groups"):
        ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
            ctx,
            NetworkInterfaceId=interface["NetworkInterfaceId"],
            Groups=new_value[device_index].get("Groups", []),
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

    # Check for changes to description
    if old_value[device_index].get("Description") != new_value[device_index].get(
        "Description"
    ):
        ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
            ctx,
            NetworkInterfaceId=interface["NetworkInterfaceId"],
            Description=new_value[device_index].get("Description", ""),
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

    # Check for changes to Delete On Termination
    if old_value[device_index].get("DeleteOnTermination") != new_value[
        device_index
    ].get("DeleteOnTermination"):
        ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
            ctx,
            NetworkInterfaceId=interface["NetworkInterfaceId"],
            Attachment=dict(
                AttachmentId=interface["Attachment"]["AttachmentId"],
                DeleteOnTermination=new_value[device_index].get("DeleteOnTermination"),
            ),
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def list_(
    hub,
    ctx,
    name: str = None,
    filters: List = None,
) -> Dict:
    """
    Use an un-managed Internet Gateway as a data-source. Supply one of the inputs as the filter.

    Args:
        name(str): The name of the Idem state.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_internet_gateways

    """
    result = dict(comment=[], ret=[], result=True)
    if filters:
        filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
            filters=filters
        )

    ret = await hub.exec.boto3.client.ec2.describe_internet_gateways(
        ctx,
        Filters=filters,
    )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    for internet_gateway in ret["ret"]["InternetGateways"]:
        resource_id = internet_gateway.get("InternetGatewayId")
        resource_converted = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_internet_gateway_to_present(
                resource=internet_gateway, idem_resource_name=resource_id
            )
        )
        result["ret"].append(resource_converted)

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util function to convert raw resource state from AWS Launch Configuration to present input format.
"""


def convert_raw_launch_configuration_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("LaunchConfigurationName")
    resource_parameters = OrderedDict(
        {
            "LaunchConfigurationName": "name",
            "ImageId": "image_id",
            "KeyName": "key_name",
            "SecurityGroups": "security_groups",
            "ClassicLinkVPCId": "classic_link_vpc_id",
            "ClassicLinkVPCSecurityGroups": "classic_link_vpc_security_groups",
            "UserData": "user_data",
            "InstanceId": "instance_id",
            "InstanceType": "instance_type",
            "KernelId": "kernel_id",
            "RamdiskId": "ramdisk_id",
            "BlockDeviceMappings": "block_device_mappings",
            "InstanceMonitoring": "instance_monitoring",
            "SpotPrice": "spot_price",
            "IamInstanceProfile": "iam_instance_profile",
            "EbsOptimized": "ebs_optimized",
            "AssociatePublicIpAddress": "associate_public_ip_address",
            "PlacementTenancy": "placement_tenancy",
            "MetadataOptions": "metadata_options",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    resource_translated = (
        hub.tool.aws.autoscaling.conversion_utils.handle_security_groups(
            resource=resource_translated
        )
    )
    return resource_translated


def handle_security_groups(hub, resource):
    # security_groups and classic_link_vpc_security_groups are of type List(str)
    # Just the change in the order of their elements should not be considered as a difference in new state and old state
    # So sorting the values for security_groups and classic_link_vpc_security_groups.
    if resource.get("security_groups"):
        resource["security_groups"] = sorted(resource["security_groups"])
    if resource.get("classic_link_vpc_security_groups"):
        resource["classic_link_vpc_security_groups"] = sorted(
            resource["classic_link_vpc_security_groups"]
        )
    return resource


def convert_raw_scaling_policy_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_id = raw_resource.get("PolicyName")

    resource_parameters = OrderedDict(
        {
            "AutoScalingGroupName": "auto_scaling_group_name",
            "PolicyType": "policy_type",
            "AdjustmentType": "adjustment_type",
            "MinAdjustmentStep": "min_adjustment_step",
            "MinAdjustmentMagnitude": "min_adjustment_magnitude",
            "ScalingAdjustment": "scaling_adjustment",
            "Cooldown": "cooldown",
            "MetricAggregationType": "metric_aggregation_type",
            "EstimatedInstanceWarmup": "estimated_instance_warmup",
            "Enabled": "enabled",
        }
    )
    resource_translated = {
        "name": raw_resource.get("PolicyName"),
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if "StepAdjustments" in raw_resource:
        resource_translated["step_adjustments"] = raw_resource.get(
            "StepAdjustments"
        ).copy()

    if "TargetTrackingConfiguration" in raw_resource:
        resource_translated["target_tracking_configuration"] = raw_resource.get(
            "TargetTrackingConfiguration"
        ).copy()

    if "PredictiveScalingConfiguration" in raw_resource:
        resource_translated["predictive_scaling_configuration"] = raw_resource.get(
            "PredictiveScalingConfiguration"
        ).copy()

    return resource_translated



""" State module for managing Elastic Container Registry (Amazon ECR) Lifecycle Policy. """
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    repository_name: str,
    lifecycle_policy_text: Dict or str,
    resource_id: str = None,
    registry_id: str = None,
) -> Dict[str, Any]:
    """Creates or updates the lifecycle policy for the specified repository. For more information, see Lifecycle policy
    template.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository. If you do not
            specify a registry, the default registry is assumed. Defaults to None.

        repository_name(str):
            The name of the repository to receive the policy.

        lifecycle_policy_text(dict or str):
            The JSON repository policy text to apply to the repository.

    Request Syntax:
        .. code-block:: sls

            [stream-name]:
              aws.ecr.lifecycle_policy.present:
                - name: "string"
                - repository_name: "string"
                - resource_id: "string"
                - registry_id: "string"
                - lifecycle_policy_text: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.ecr.lifecycle_policy.present:
                - name: idem_test_lifecycle_policy
                - repository_name: idem_test_ecr_repository
                - lifecycle_policy_text: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    lifecycle_policy_text = hub.tool.aws.state_comparison_utils.standardise_json(
        lifecycle_policy_text
    )

    if resource_id:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

        before = await hub.exec.aws.ecr.lifecycle_policy.get(
            ctx,
            name=name,
            resource_id=resource_id,
            registry_id=registry_id,
            repository_name=repository_name,
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = tuple(before["comment"])
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.ecr.lifecycle_policy", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        result["new_state"] = copy.deepcopy(before["ret"])

        if (
            lifecycle_policy_text
            and not hub.tool.aws.state_comparison_utils.is_json_identical(
                result["old_state"]["lifecycle_policy_text"], lifecycle_policy_text
            )
        ):
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.ecr.lifecycle_policy", name=name
                )
                result["new_state"]["lifecycle_policy_text"] = lifecycle_policy_text
                return result

            update_return = await hub.exec.boto3.client.ecr.put_lifecycle_policy(
                ctx=ctx,
                registryId=registry_id,
                repositoryName=repository_name,
                lifecyclePolicyText=lifecycle_policy_text,
            )
            result["comment"] = tuple(update_return["comment"])
            if not update_return["result"]:
                result["result"] = False
                return result
            resource_updated = bool(update_return["ret"])
            if resource_updated:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.ecr.lifecycle_policy", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "repository_name": repository_name,
                    "lifecycle_policy_text": lifecycle_policy_text,
                    "registry_id": registry_id,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ecr.lifecycle_policy", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ecr.put_lifecycle_policy(
            ctx,
            registryId=registry_id,
            repositoryName=repository_name,
            lifecyclePolicyText=lifecycle_policy_text,
        )
        if not ret["result"] and not ret["ret"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ecr.lifecycle_policy", name=name
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ecr.lifecycle_policy.get(
            ctx,
            name=name,
            resource_id=resource_id,
            registry_id=registry_id,
            repository_name=repository_name,
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += tuple(after["comment"])
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    repository_name: str = None,
    resource_id: str = None,
    registry_id: str = None,
) -> Dict[str, Any]:
    """Deletes the lifecycle policy associated with the specified repository.

    Args:
        name(str):
            An Idem name of the resource.

        repository_name(str):
            The name of the ECR repository in Amazon Web Services that contains the policy to delete.

        resource_id(str, Optional):
            The registry id and repository name with a separator '-'. Format: ``[registry_id]-[repository_name]``.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository. If you do not
            specify a registry, the default registry is assumed. Defaults to None.

    Request Syntax:
      .. code-block:: sls

           [lifecycle_policy-name]:
              aws.ecr.lifecycle_policy.absent:
                - name: 'string'
                - repository_name: 'string'
                - resource_id: 'string'
                - registry_id: 'string'

    Returns:
      Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ecr.lifecycle_policy.absent:
                - name: idem_test_policy
                - repository_name: idem_test_repository
                - resource_id: idem_test_ecr_registry-idem_test_ecr_repository
                - registry_id: idem_test_registry
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.lifecycle_policy", name=name
        )
        return result
    else:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]
            repository_name = resource_id.split("-")[1]

    before = await hub.exec.aws.ecr.lifecycle_policy.get(
        ctx,
        name=name,
        resource_id=resource_id,
        registry_id=registry_id,
        repository_name=repository_name,
    )
    if not before["result"]:
        result["comment"] = tuple(before["comment"])
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.lifecycle_policy", name=name
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ecr.lifecycle_policy", name=name
            )
            return result
        else:
            ret = await hub.exec.boto3.client.ecr.delete_lifecycle_policy(
                ctx,
                registryId=registry_id,
                repositoryName=repository_name,
            )
            if not ret["result"]:
                result["comment"] = before["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ecr.lifecycle_policy", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Pass required params to get an Amazon Elastic Container Registry (Amazon ECR) Lifecycle Policy resource.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.
    To describe all the repository policies, we first need to list all the repositories, and then we get the policy
    associated to each repository

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ecr.lifecycle_policy
    """
    result = {}
    describe_ret = await hub.exec.aws.ecr.repository.list(ctx)
    if not describe_ret["result"]:
        hub.log.warning(f"Could not describe repositories: {describe_ret['comment']}")
        return result

    for repository in describe_ret["ret"]:
        registry_id = repository.get("registry_id")
        repository_name = repository.get("name")

        policy_ret = await hub.exec.aws.ecr.lifecycle_policy.get(
            ctx,
            name=repository_name,
            resource_id=(f"{registry_id}-{repository_name}"),
            registry_id=registry_id,
            repository_name=repository_name,
        )

        if not policy_ret["result"]:
            hub.log.warning(
                f"Could not get lifecycle policy for repository '{registry_id}/{repository_name}': "
                f"{policy_ret['comment']}. Describe will skip this lifecycle policy and continue."
            )
            continue

        if not policy_ret.get("ret"):
            hub.log.debug(
                f"No lifecycle policy for repository '{registry_id}/{repository_name}' exists: "
                f"{policy_ret['comment']}. Describe will skip this lifecycle policy and continue."
            )
            continue

        resource = policy_ret.get("ret")
        result[resource.get("resource_id")] = {
            "aws.ecr.lifecycle_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_rest_api_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """Convert AWS returned data structure to correct idem apigateway.rest_api present state

     Args:
        raw_resource: The aws response to convert

        idem_resource_name (str, Optional): An idem name of the resource

    Returns: Valid idem state for apigateway.rest_api of type dict['string', Any]
    """
    resource_id = raw_resource.get("id")
    resource_parameters = OrderedDict(
        {
            "description": "description",
            "version": "version",
            "binaryMediaTypes": "binary_media_types",
            "minimumCompressionSize": "minimum_compression_size",
            "apiKeySource": "api_key_source",
            "endpointConfiguration": "endpoint_configuration",
            "policy": "policy",
            "tags": "tags",
            "disableExecuteApiEndpoint": "disable_execute_api_endpoint",
            "rootResourceId": "root_resource_id",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update_rest_api(
    hub, ctx, old_state: Dict[str, Any], updatable_resource_parameters: Dict[str, Any]
):
    """Updates an AWS API Gateway API resource.

    Args:
        old_state (dict): Previous state of the rest_api resource

        updatable_resource_parameters (dict): Parameters from SLS file.

    Returns:
        dict[str, Any]
    """
    result = dict(comment=[], result=True, ret=None)
    patch_operations = []

    new_state = copy.deepcopy(old_state)
    resource_updated = False
    for key, value in updatable_resource_parameters.items():
        if value is not None:
            if key == "endpoint_configuration":
                if len(value["types"]) != len(old_state[key]["types"]):
                    # there can be multiple types at once, but add/remove aren't supported.
                    # see https://docs.aws.amazon.com/apigateway/latest/api/patch-operations.html
                    result["comment"] += (
                        f"AWS doesn't support changing the number of endpoint types. You are trying to go from "
                        f"{value['types']} to {old_state[key]['types']}. The lengths must be "
                        f"equal; {len(value['types'])} != {len(old_state[key]['types'])}.",
                    )
                    result["result"] = False
                    return result
                for i in range(len(value["types"])):
                    if value["types"][i] != old_state[key]["types"][i]:
                        patch_operations.append(
                            {
                                "op": "replace",
                                "path": f"/endpointConfiguration/types/{old_state[key]['types'][i]}",
                                "value": value["types"][i],
                            }
                        )
                        new_state[key]["types"][i] = value["types"][i]
                        resource_updated = True

                old_vpc_endpoint_ids = old_state["endpoint_configuration"].get(
                    "vpcEndpointIds", []
                )
                new_vpc_endpoint_ids = value.get("vpcEndpointIds", [])
                if not (
                    hub.tool.aws.state_comparison_utils.are_lists_identical(
                        old_vpc_endpoint_ids,
                        new_vpc_endpoint_ids,
                    )
                ):
                    endpoints_to_remove = list(
                        set(old_vpc_endpoint_ids).difference(set(new_vpc_endpoint_ids))
                    )
                    endpoints_to_add = list(
                        set(new_vpc_endpoint_ids).difference(set(old_vpc_endpoint_ids))
                    )
                    if endpoints_to_remove:
                        patch_operations.append(
                            {
                                "op": "remove",
                                "path": f"/endpointConfiguration/vpcEndpointIds",
                                "value": ",".join(endpoints_to_remove),
                            }
                        )
                    if endpoints_to_add:
                        patch_operations.append(
                            {
                                "op": "add",
                                "path": f"/endpointConfiguration/vpcEndpointIds",
                                "value": ",".join(endpoints_to_add),
                            }
                        )

                    new_state["endpoint_configuration"][
                        "vpcEndpointIds"
                    ] = new_vpc_endpoint_ids
                    resource_updated = True
            elif (
                isinstance(value, list)
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    value,
                    old_state.get(key, []),
                )
                or value != old_state[key]
            ):
                if old_state.get(key):
                    patch_operations.append(
                        {
                            "op": "replace",
                            "path": f"/{key}",
                            "value": str(value)
                            if not isinstance(value, list)
                            else ",".join(value),
                        }
                    )
                else:
                    patch_operations.append(
                        {
                            "op": "add",
                            "path": f"/{key}",
                            "value": str(value)
                            if not isinstance(value, list)
                            else ",".join(value),
                        }
                    )
                new_state[key] = value
                resource_updated = True
    if ctx.get("test", False):
        if resource_updated:
            result["ret"] = new_state
        return result

    if resource_updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_rest_api(
            ctx, restApiId=old_state["resource_id"], patchOperations=patch_operations
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] += update_ret["comment"]
            return result
        else:
            result["comment"] += hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.rest_api", name=old_state["name"]
            )
            result["ret"] = new_state

    else:
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.rest_api", name=old_state["name"]
        )

    return result



"""Exec module for managing Amazon DocDB DB Cluster Parameter Group."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}

RESOURCE_TYPE = "aws.docdb.db_cluster_parameter_group"


async def get(hub, ctx, name: str, resource_id: str = None):
    """Retrieves the specified AWS Document DB Cluster Parameter Group.

    Args:
        resource_id(str, Optional):
            AWS Document DB Cluster Parameter Group Name.
        name(str):
            An Idem name of the resource.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.docdb.db_cluster_parameter_group.get resource_id="resource_id" name="name"

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.docdb.db_cluster_parameter_group.get(
                    ctx, name=name, resource_id=resource_id
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.docdb.db_cluster_parameter_group.get
                - kwargs:
                    name: name
                    resource_id: resource_id

    """
    result = dict(comment=[], ret=None, result=True)
    before = await hub.tool.aws.docdb.db_cluster_parameter_group.search_raw(
        ctx, resource_id=resource_id
    )
    if not before["result"]:
        if "DBParameterGroupNotFoundFault" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type=RESOURCE_TYPE,
                    name=name,
                )
            )
            result["comment"].append(before["comment"])
            return result
        result["result"] = False
        result["comment"].append(before["comment"])
        return result
    if not before["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc", name=name
            )
        )
        return result

    resource = before["ret"]["DBClusterParameterGroups"]
    if len(resource) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id, resource_type=RESOURCE_TYPE
            )
        )
    resource_arn = before["ret"]["DBClusterParameterGroups"][0].get(
        "DBClusterParameterGroupArn"
    )

    tags = await hub.exec.boto3.client.docdb.list_tags_for_resource(
        ctx, ResourceName=resource_arn
    )
    if not tags["result"]:
        result["result"] = False
        result["comment"].append(tags["comment"])
        return result

    parameters_raw = await hub.exec.boto3.client.docdb.describe_db_cluster_parameters(
        ctx=ctx, DBClusterParameterGroupName=resource_id
    )
    if not parameters_raw["result"]:
        result["result"] = False
        result["comment"].append(parameters_raw["comment"])
        return result
    parameters_raw = parameters_raw.get("ret")
    parameters = parameters_raw.get("Parameters")

    result[
        "ret"
    ] = hub.tool.aws.docdb.db_cluster_parameter_group.convert_raw_db_cluster_parameter_group_to_present(
        raw_resource=before["ret"]["DBClusterParameterGroups"][0],
        idem_resource_name=resource_id,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags["ret"]["TagList"]),
        parameters=parameters,
    )
    return result


async def list_(hub, ctx, name: str = None, filters: List = None) -> Dict:
    """Use a list of un-managed docdb db cluster parameter groups as a data-source.

    Args:
        name(str, Optional):
            The name of idem state.

        filters(list, Optional):
            One or more filters, dictionary of attribute value pair which will be converted to boto3 filter
            format.

    .. note::
        At the time of writing this module, filters parameter is not supported for this resource as per
        AWS documentation so idem-aws has omitted it.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS DocDB DB Cluster Parameter Group in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.docdb.db_cluster_parameter_group.list name=name

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, **kwargs):
                ret = await hub.exec.aws.docdb.db_cluster_parameter_group.list(
                    ctx, name=name
                )

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.docdb.db_cluster_parameter_group.list
                - kwargs:
                    name: 'test-doc-db-cluster-param-group'


    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.docdb.db_cluster_parameter_group.search_raw(
        ctx=ctx, filters=filters
    )
    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result
    for db_cluster_parameter_group in ret["ret"]["DBClusterParameterGroups"]:
        db_cluster_parameter_group_name = db_cluster_parameter_group.get(
            "DBClusterParameterGroupName"
        )
        arn = db_cluster_parameter_group["DBClusterParameterGroupArn"]

        # get tags
        tags = await hub.tool.aws.docdb.tag.get_tags_for_resource(
            ctx=ctx, resource_arn=arn
        )
        if not tags["result"]:
            result["result"] = False
            result["comment"].append(tags["comment"])
            return result
        tags = tags["ret"]

        parameters_raw = (
            await hub.exec.boto3.client.docdb.describe_db_cluster_parameters(
                ctx=ctx, DBClusterParameterGroupName=db_cluster_parameter_group_name
            )
        )
        if not parameters_raw["result"]:
            result["result"] = False
            result["comment"].append(parameters_raw["comment"])
            return result
        parameters_raw = parameters_raw.get("ret")
        parameters = parameters_raw.get("Parameters")

        result["ret"].append(
            hub.tool.aws.docdb.db_cluster_parameter_group.convert_raw_db_cluster_parameter_group_to_present(
                raw_resource=db_cluster_parameter_group,
                idem_resource_name=db_cluster_parameter_group_name,
                tags=tags,
                parameters=parameters,
            )
        )
    return result



"""Exec module for managing Amazon EC2 Container Registry (ECR) Lifecycle Policies."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    registry_id: str = None,
    repository_name: str = None,
) -> Dict[str, Any]:
    """Retrieves a Amazon EC2 Container Registry (ECR) Lifecycle Policy created on a repository in a registry.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The name of the ECR repository in Amazon Web Services. If resource_id is not None, it takes precedence
            in get API call.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository
            to search. If you do not specify a registry, the default registry is assumed.

        repository_name(str):
            The name of the repository to receive the policy.

    Returns:
        Dict[str, Any]:
            Return image repository in a given registry.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.lifecycle_policy.get name='idem_name' resource_id='ecr_lifecycle_policy'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.lifecycle_policy.get
                - kwargs:
                    name: 'idem_name'
                    resource_id: 'ecr_lifecycle_policy'
    """
    result = dict(comment=[], ret=None, result=True)

    if resource_id:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

        if repository_name is None:
            repository_name = resource_id.split("-")[1]

    describe_ret = await hub.exec.boto3.client.ecr.get_lifecycle_policy(
        ctx=ctx, registryId=registry_id, repositoryName=repository_name
    )

    if not describe_ret["result"]:
        # Do not return success=false when it is not found.
        if "RepositoryNotFoundException" in str(
            describe_ret["comment"]
        ) or "LifecyclePolicyNotFoundException" in str(describe_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ecr.lifecycle_policy", name=resource_id
                )
            )
            result["comment"] += list(describe_ret["comment"])
            return result

        result["comment"] += list(describe_ret["comment"])
        result["result"] = False
        return result

    if not describe_ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ecr.lifecycle_policy", name=resource_id
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_lifecycle_policy_to_present(
        raw_resource=describe_ret.get("ret"),
        idem_resource_name=name,
    )
    return result



"""Util functions for AWS API Gateway v2 tags."""
import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
) -> Dict[str, Any]:
    """Updates tags for an AWS API Gateway v2 resource.

    Args:
        resource_arn(str):
            AWS API Gateway v2 resource arn.

        old_tags(dict[str, Any]):
            List of existing tags in the format of ``{"Key": tag-key, "Value": tag-value}``.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

        new_tags(dict[str, Any]):
            List of new tags in the format of ``{"Key": tag-key, "Value": tag-value}``.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

    Returns:
        Dict[str, Any]:
            Returns updated tags for the resource.
    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = []
    tags_result = copy.deepcopy(old_tags)

    for key, value in new_tags.items():
        if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
            key not in old_tags
        ):
            tags_to_add[key] = value

    for key in old_tags:
        if key not in new_tags:
            tags_to_remove.append(key)

    if tags_to_remove:
        if not ctx.get("test", False):
            untag_ret = await hub.exec.boto3.client.apigatewayv2.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not untag_ret["result"]:
                result["result"] = False
                result["comment"] = untag_ret["comment"]
                return result
        [tags_result.pop(key, None) for key in tags_to_remove]

    if tags_to_add:
        if not ctx.get("test", False):
            tag_ret = await hub.exec.boto3.client.apigatewayv2.tag_resource(
                ctx, ResourceArn=resource_arn, Tags=tags_to_add
            )
            if not tag_ret["result"]:
                result["result"] = False
                result["comment"] = tag_ret["comment"]
                return result

    if ctx.get("test", False):
        result["comment"] = (
            f"Would update tags: Add {list(tags_to_add.keys())} Remove {tags_to_remove}",
        )
    else:
        result["comment"] = (
            f"Updated tags: Add {list(tags_to_add.keys())} Remove {tags_to_remove}",
        )
    result["ret"] = {"tags": {**tags_result, **tags_to_add}}
    return result



"""Exec module for managing AWS SES Identity notification topic."""

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str):
    """Get SES Notification Topic from AWS account.

    Args:
        name(str):
            A name for the Identity Notification Topic.

        resource_id (str):
            The identifier for AWS Identity Notification Topic, combination of identity, notification topic separated by
            a separator '::'.

    Returns:
        Dict[str, Any]:
            Returns Notification Topic in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.ses.identity_notification_topic.get name = "my_resource" resource_id="example.com::Complaint"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ses.identity_notification_topic.get
                - kwargs:
                    name: "my_resource"
                    resource_id:"example.com::Complaint"
    """
    result = dict(comment=[], ret=None, result=True)

    identity = resource_id.split("::")[0]

    ret = await hub.exec.aws.ses.identity_notification_topic.list(ctx, [identity])

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"]:
        result["comment"] += (
            f"No identity notification set for identity '{identity}'",
        )

    for resource in ret["ret"]:
        if resource.get("resource_id") == resource_id:
            result["ret"] = resource
            break

    return result


async def list_(hub, ctx, identities: list = None):
    """Lists SES Domain notification topics.

    Args:
        name (str):
            A name for the Identity Notification Topic.

        identities (list):
            List of AWS Identity Notification Topic IDs.

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.ses.identity_notification_topic.list identities="example.com"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ses.identity_notification_topic.list
                - kwargs:
                    identities:"example.com"

    """
    result = dict(comment=[], ret=[], result=True)

    if not identities:
        ret_ids = await hub.exec.boto3.client.ses.list_identities(ctx)

        if not ret_ids["result"]:
            result["comment"] += ret_ids["comment"]
            result["result"] = False
            return result

        identities = ret_ids["ret"].get("Identities")

    ret = await hub.exec.boto3.client.ses.get_identity_notification_attributes(
        ctx, Identities=identities
    )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    for identity, attributes in ret["ret"]["NotificationAttributes"].items():

        if "BounceTopic" in attributes:
            result["ret"].append(
                hub.tool.aws.ses.conversion_utils.convert_raw_identity_notification_topic_to_present(
                    attributes=attributes,
                    identity=identity,
                    notification_type="Bounce",
                )
            )

        if "ComplaintTopic" in attributes:
            result["ret"].append(
                hub.tool.aws.ses.conversion_utils.convert_raw_identity_notification_topic_to_present(
                    attributes=attributes,
                    identity=identity,
                    notification_type="Complaint",
                )
            )

        if "DeliveryTopic" in attributes:
            result["ret"].append(
                hub.tool.aws.ses.conversion_utils.convert_raw_identity_notification_topic_to_present(
                    attributes=attributes,
                    identity=identity,
                    notification_type="Delivery",
                )
            )

    return result



from typing import Dict


async def get(hub, ctx, name: str, resource_id: str = None) -> Dict:
    """Pass required params to get a Kinesis stream resource. Retrieves the specified Kinesis Stream.

    Args:
        name(str):
          AWS Kinesis Stream name to uniquely identify the resource.

        resource_id(str, Optional):
            AWS Kinesis Stream name.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The Kinesis Stream in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.kinesis.stream.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.kinesis.stream.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.kinesis.stream.search_raw(
        ctx=ctx,
        name=resource_id if resource_id else name,
    )
    if not ret["result"]:
        if "ResourceNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.kinesis.stream", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if "StreamDescriptionSummary" not in ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.kinesis.stream", name=name
            )
        )
        return result

    resource_id = ret["ret"]["StreamDescriptionSummary"].get("StreamName")
    tags = {}
    tags_ret = await hub.exec.boto3.client.kinesis.list_tags_for_stream(
        ctx, StreamName=resource_id if resource_id else name
    )
    if not tags_ret["result"]:
        result["comment"] = list(tags_ret["comment"])
        result["result"] = False
        return result
    else:
        if tags_ret["ret"] and tags_ret.get("ret")["Tags"]:
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                tags_ret.get("ret")["Tags"]
            )

    result["ret"] = hub.tool.aws.kinesis.conversion_utils.convert_raw_stream_to_present(
        raw_resource=ret["ret"]["StreamDescriptionSummary"],
        idem_resource_name=resource_id,
        tags=tags,
    )
    return result



"""State module for managing SES Domain Identity."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    domain: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Adds a domain to the list of identities for your Amazon SES account in the current AWS Region and attempts to verify it.

    Args:
        name(str):
            An Idem name to identify SES Domain Identity resource.

        domain(str):
            The domain to be verified.

        resource_id(str, Optional):
            The domain name.

    Request Syntax:
        .. code-block:: sls

            [domain_identity-resource-id]:
              aws.ses.domain_identity.present:
                - name: "string"
                - domain: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            example.com:
              aws.ses.domain_identity.present:
                - name: example.com
                - domain: example.com

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    if resource_id:
        before = await hub.exec.aws.ses.domain_identity.get(
            ctx, resource_id=resource_id
        )
        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = before["result"]
            return result

    if before and before.get("ret"):
        result["old_state"] = before["ret"].copy()
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.ses.domain_identity", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "domain": domain,
                },
            )
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ses.domain_identity", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ses.verify_domain_identity(
            ctx,
            Domain=domain,
        )
        if not ret["result"]:
            result["comment"] += ret["comment"]
            result["result"] = False
            return result

        after = await hub.exec.aws.ses.domain_identity.get(ctx, resource_id=domain)
        result["comment"] += hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ses.domain_identity", name=name
        )
        result["new_state"] = after.get("ret")
        if not ret["result"]:
            result["comment"] += ret["comment"]
            result["result"] = False
            return result

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified identity

    Args:
        name(str):
            An Idem name to identify SES Domain Identity resource.

        resource_id(str, Optional):
            The domain name.
            Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [domain_identity-resource-id]:
              aws.ses.domain_identity.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            example.com:
              aws.ses.domain_identity.absent:
                - name: example.com
                - domain: example.com
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ses.domain_identity", name=name
        )
        return result

    before = await hub.exec.aws.ses.domain_identity.get(ctx, resource_id)
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ses.domain_identity", name=name
        )
    else:
        result["old_state"] = before["ret"]

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ses.domain_identity", name=name
            )
            return result
        # delete the SES Domain Identity

        response = await hub.exec.boto3.client.ses.delete_identity(
            ctx,
            Identity=resource_id,
        )
        if not response["result"]:
            result["comment"] = (
                f"Error deleting aws.ses.domain_identity '{resource_id}': {response['comment']}",
            )
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ses.domain_identity", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            idem describe aws.ses.domain_identity
    """

    result = {}

    ret = await hub.exec.aws.ses.domain_identity.list(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    for identity in ret["ret"]:

        resource_id = identity.get("resource_id")

        result[resource_id] = {
            "aws.ses.domain_identity.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in identity.items()
            ]
        }

    return result



"""This module defines the present, absent and describe functionality for AWS document db cluster parameter group"""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

RESOURCE_TYPE = "aws.docdb.db_cluster_parameter_group"


async def present(
    hub,
    ctx,
    name: str,
    db_parameter_group_family: str,
    description: str,
    resource_id: str = None,
    parameters: List[
        make_dataclass(
            """DB Parameter group configuration.""" "Parameter",
            [
                ("ParameterName", str),
                ("ParameterValue", str),
                ("ApplyMethod", str, field(default=None)),
                ("Description", str, field(default=None)),
                ("Source", str, field(default=None)),
                ("ApplyType", str, field(default=None)),
                ("DataType", str, field(default=None)),
                ("AllowedValues", str, field(default=None)),
                ("MinimumEngineVersion", str, field(default=None)),
                ("SupportedEngineModes", List[str], field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Creates a new cluster parameter group.

    Parameters in a cluster parameter group apply to all of the instances in a cluster.

    A cluster parameter group is initially created with the default parameters for the database engine used by instances
    in the cluster. In Amazon DocumentDB, you cannot make modifications directly to the default.docdb3.6 cluster
    parameter group. If your Amazon DocumentDB cluster is using the default cluster parameter group and you want to
    modify a value in it, you must first create a new parameter group or copy an existing parameter group , modify it,
    and then apply the modified parameter group to your cluster. For the new cluster parameter group and associated
    settings to take effect, you must then reboot the instances in the cluster without failover. For more information,
    see Modifying Amazon DocumentDB Cluster Parameter Groups.

    Args:
        name(str):
            DBClusterParameterGroupName parameter on the resource. This value is stored as a lowercase string.
            This is also used as idem name.
        db_parameter_group_family(str):
            The cluster parameter group family name.
        description(str):
            The description for the DB cluster parameter group.
        resource_id(str, Optional):
            DBClusterParameterGroupName of the resource. Defaults to None.
        parameters(list, Optional): parameters in the DB cluster parameter group. Defaults to None.
            These parameters can only be modified.
            To modify more than one parameter, submit a list of the following: ParameterName, ParameterValue, and
            ApplyMethod. A maximum of 20 parameters can be modified in a single request.

            .. note::
                Changes to dynamic parameters are applied immediately. Changes to static parameters require a reboot
                without failover to the DB cluster associated with the parameter group before the change can take
                effect.

            .. warning::
                After you create a cluster parameter group, you should wait at least 5 minutes before creating your
                first cluster that uses that cluster parameter group as the default parameter group. This allows
                Amazon DocumentDB to fully complete the create action before the parameter group is used as the default
                for a new cluster. This step is especially important for parameters that are critical when creating the
                default database for a cluster, such as the character set for the default database defined by the
                character_set_database parameter.

            Below are the attributes available for parameters:

            Type: object dict(Describes a parameter)

            * ParameterName (str):
                Specifies the name of the parameter.
            * ParameterValue (str):
                Specifies the value of the parameter.
            * Description (str, Optional):
                provides a description of the parameter.
            * Source (str, Optional):
                Indicates the source of the parameter value.
            * ApplyType (str, Optional):
                Specifies the engine specific parameters type.
            * DataType (str, Optional):
                Specifies the valid data type for the parameter.
            * AllowedValues (str, Optional):
                Specifies the valid range of values for the parameter.
            * MinimumEngineVersion (str, Optional):
                The earliest engine version to which the parameter can apply.
            * ApplyMethod (str, Optional):
                Indicates when to apply parameter updates.
            * SupportedEngineModes (list, Optional):
                The valid DB engine modes.
        tags(dict, Optional):
            The tags to be assigned to the new DB cluster parameter group. Defaults to None.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_docdb_db_cluster_parameter_group]:
              aws.docdb.db_cluster_parameter_group.present:
              - db_cluster_parameter_group_arn: 'string'
              - db_parameter_group_family: 'string'
              - description: 'string'
              - name: 'string'
              - parameters:
                - AllowedValues: 'string'
                  ApplyMethod: 'string'
                  ApplyType: 'string'
                  DataType: 'string'
                  Description: 'string'
                  IsModifiable: 'string'
                  ParameterName: 'string'
                  Source: 'string'


    Examples:
        .. code-block:: sls

            default.aurora5.6:
              aws.docdb.db_cluster_parameter_group.present:
              - db_cluster_parameter_group_arn: arn:aws:rds:ca-central-1:746014882121:cluster-pg:default.aurora5.6
              -  db_parameter_group_family: aurora5.6
              -  description: Default cluster parameter group for aurora5.6
              -  name: default.aurora5.6
              -  parameters:
                - AllowedValues: 0,1
                  ApplyMethod: pending-reboot
                  ApplyType: static
                  DataType: boolean
                  Description: Controls whether user-defined functions that have only an xxx symbol
                    for the main function can be loaded
                  IsModifiable: false
                  ParameterName: allow-suspicious-udfs
                  Source: engine-default

    """

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    desired_state = {
        "name": name,
        "resource_id": name,
        "db_parameter_group_family": db_parameter_group_family,
        "description": description,
        "parameters": parameters,
        "tags": tags,
    }
    is_created = False

    if not resource_id:
        # if there is no resource_id, it is expected that we create the resource
        # if this is test run, just generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["resource_id"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        else:
            # if this is not a test run, create the real resource
            create_db_cluster_parameter_group_result = (
                await hub.exec.boto3.client.docdb.create_db_cluster_parameter_group(
                    ctx,
                    DBClusterParameterGroupName=name,
                    DBParameterGroupFamily=db_parameter_group_family,
                    Description=description,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )

            # if create failed, fail the request
            if not create_db_cluster_parameter_group_result["result"]:
                result["comment"] = create_db_cluster_parameter_group_result["comment"]
                result["resul"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
            resource_id = create_db_cluster_parameter_group_result["ret"][
                "DBClusterParameterGroup"
            ]["DBClusterParameterGroupName"]
        is_created = True

    update_db_cluster_parameter_group_result = (
        await hub.tool.aws.docdb.db_cluster_parameter_group.update(
            ctx,
            name=name,
            resource_id=resource_id,
            desired_state=desired_state,
            is_created=is_created,
        )
    )

    if not (
        update_db_cluster_parameter_group_result
        and update_db_cluster_parameter_group_result["result"]
        and update_db_cluster_parameter_group_result["new_state"]
    ):
        result["result"] = False
        result["comment"] = update_db_cluster_parameter_group_result["comment"]
        return result

    result["old_state"] = update_db_cluster_parameter_group_result["old_state"]
    result["new_state"] = update_db_cluster_parameter_group_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] = update_db_cluster_parameter_group_result["comment"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """

    Deletes a specified cluster parameter group. The cluster parameter group to be deleted can't be associated with any clusters.

    Args:
        name(str):
            DBClusterParameterGroupName parameter on the resource.
            This is also used as idem name.
        resource_id(str, Optional):
            DBClusterParameterGroupName of the resource. Defaults to None.
            Idem automatically considers the resource being absent if resource_id is not specified.

            Constraints: Must match the name of an existing DBClusterParameterGroupName.

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.docdb.db_cluster_parameter_group.absent:
                - name: 'test-db-clus-param-group'
                - resource_id: 'test-db-clus-param-group'

    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type=RESOURCE_TYPE, name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.docdb.db_cluster_parameter_group.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        return result
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.docdb.delete_db_cluster_parameter_group(
            ctx,
            DBClusterParameterGroupName=resource_id,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns a list of DBClusterParameterGroup.

    Examples:
        .. code-block:: bash

            $ idem describe aws.docdb.db_cluster_parameter_group

    """
    result = {}
    ret = await hub.exec.aws.docdb.db_cluster_parameter_group.list(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe {RESOURCE_TYPE} {ret['comment']}")
        return result

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "aws.docdb.db_cluster_parameter_group.present": [resource]
        }
    return result



__func_alias__ = {"list_": "list"}


async def list_(hub, ctx):
    """Lists AWS IAM Users.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The IAM Policies in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.iam.user.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.iam.user.list
    """
    ret = await hub.exec.boto3.client.iam.list_users(ctx)

    if not ret["result"]:
        return {
            "result": False,
            "ret": [],
            "comment": f"Can not list users: {ret['comment']}",
        }

    users = []
    for user_raw in ret["ret"]["Users"]:
        # we're using this function for the camel case to snake case
        user = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(user_raw)
        users.append(user)

    return {"result": True, "ret": users}



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_method_to_present(
    hub,
    raw_resource: Any,
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """Convert AWS API Gateway Method to a common idem present state.

    Args:
        resource_id(str): Idem Resource ID that is generated once the resource is created.

        raw_resource(dict[str, Any]): The AWS response to convert.

        idem_resource_name(str, Optional): The idem name of the resource.

    Returns:
        dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "httpMethod": "http_method",
            "authorizationType": "authorization_type",
            "apiKeyRequired": "api_key_required",
        }
    )

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": resource_id.split("-")[0],
        "parent_resource_id": resource_id.split("-")[1],
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if idem_resource_name:
            resource_translated["name"] = idem_resource_name
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    update_parameters: Dict[str, Any],
):
    """Updates an API Gateway Method

    Args:
        old_state(dict): Previous state of resource

        update_parameters(dict): Parameters from SLS File

    Returns:
        dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in update_parameters.items():
        if key == "api_key_required":
            key_final = "apiKeyRequired"
        elif key == "authorization_type":
            key_final = "authorizationType"
        elif key == "http_method":
            key_final = "httpMethod"

        if value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            updated = True
        elif old_state.get(key) is None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_method(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["parent_resource_id"],
            httpMethod=old_state["http_method"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.method", name=old_state["name"]
        )

    return result



"""Exec module for AWS API Gateway Domain Name resources."""
__func_alias__ = {"list_": "list"}

from typing import Dict


async def get(hub, ctx, name: str, resource_id: str) -> Dict:
    """
    Get an API Gateway domain name resource from AWS with the domain name as the resource_id.

    Args:
        name(str):
            The name of the Idem state domain name.
        resource_id(str):
            AWS API Gateway domain name.

    Returns:
        Dict[str, Any]

    Examples:

        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.apigateway.domain_name.get name="unmanaged_domain_names"


        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigateway.domain_name.get
                - kwargs:
                    name: unmanaged_domain_name
                    resource_id: resource_id

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.apigateway.get_domain_name(
        ctx=ctx, domainName=resource_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.domain_name", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = await hub.tool.aws.apigateway.domain_name.convert_raw_domain_name_to_present(
        raw_resource=ret["ret"], idem_resource_name=name
    )
    return result


async def list_(hub, ctx, name: str = None):
    """Get the list of domain names for AWS APIGateway.

    Args:
        name (str, Optional):
            The name of the Idem state for logging.

    Returns:
        Dict[str, Any]

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            idem exec aws.apigateway.domain_name.list

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigateway.domain_name.list
                - kwargs:
                    name: my-resource-name
    """
    result = dict(comment=[], ret=[], result=True)
    get_domain_names_ret = await hub.exec.boto3.client.apigateway.get_domain_names(ctx)
    if (
        "NotFoundException" in str(get_domain_names_ret["comment"])
        or get_domain_names_ret["ret"]["items"] == []
    ):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.apigateway.domain_name", name="domain_name_resource"
            )
        )
        return []
    for domain_name in get_domain_names_ret["ret"]["items"]:
        idem_resource_name = domain_name["domainName"]
        get_translated_resource = await hub.tool.aws.apigateway.domain_name.convert_raw_domain_name_to_present(
            raw_resource=domain_name, idem_resource_name=idem_resource_name
        )
        result["ret"].append(get_translated_resource)
    return result



"""State module for managing AWS Doc DB cluster."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
create_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "creating",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]
update_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "modifying",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "resetting-master-credentials",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "renaming",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "upgrading",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]
delete_waiter_acceptors = [
    {
        "matcher": "error",
        "expected": "DBClusterNotFoundFault",
        "state": "success",
        "argument": "Error.Code",
    },
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "deleting",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    engine: str,
    availability_zones: List = None,
    backup_retention_period: int = None,
    db_cluster_parameter_group_name: str = None,
    vpc_security_group_ids: List = None,
    db_subnet_group_name: str = None,
    engine_version: str = None,
    port: int = None,
    master_username: str = None,
    master_user_password: str = None,
    preferred_backup_window: str = None,
    preferred_maintenance_window: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
    storage_encrypted: bool = None,
    kms_key_id: str = None,
    enable_cloudwatch_logs_exports: List = None,
    deletion_protection: bool = True,
    source_region: str = None,
    apply_immediately: bool = None,
    cloudwatch_logs_export_configuration: make_dataclass(
        """Cloudwatch log export configuration.""" "CloudwatchLogsExportConfiguration",
        [
            ("EnableLogTypes", List[str], field(default=None)),
            ("DisableLogTypes", List[str], field(default=None)),
        ],
    ) = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates a new Amazon DocumentDB cluster.

    Args:
        name(str):
            An Idem name of the resource. This also maps to db_cluster_identifier while creating the cluster.
        resource_id(str, Optional):
            An identifier of the resource in the provider, this also maps to DBClusterIdentifier during update. Defaults to None.
            The DB cluster identifier - This parameter is stored as a lowercase string. Constraints: Must
            contain from 1 to 63 letters, numbers, or hyphens. First character must be a letter. Cannot
            end with a hyphen or contain two consecutive hyphens. Example: my-cluster1.
        availability_zones(list, Optional):
            A list of EC2 Availability Zones that instances in the DB cluster can be created in. Defaults to None.
        backup_retention_period(int, Optional):
            The number of days for which automated backups are retained. You must specify a minimum value of
            1. Default: 1 Constraints: Must be a value from 1 to 35. Defaults to None.
        db_cluster_parameter_group_name(str, Optional):
            The name of the DB cluster parameter group to associate with this DB cluster. If this argument
            is omitted, the default is used. Constraints: If supplied, must match the name of an existing
            DBClusterParameterGroup. Defaults to None.
        vpc_security_group_ids(list, Optional):
            A list of EC2 VPC security groups to associate with this DB cluster. Defaults to None.
        db_subnet_group_name(str, Optional):
            A DB subnet group to associate with this DB cluster. Constraints: Must match the name of an
            existing DBSubnetGroup. Must not be default. Example: mySubnetgroup. Defaults to None.
        engine(str):
            The name of the database engine to be used for this DB cluster. Valid Values: docdb.
        engine_version(str, Optional):
            The version number of the database engine to use for the new DB cluster. Example: 1.0.2.1. Defaults to
            default to the latest major engine version.
        port(int, Optional):
            The port number on which the instances in the DB cluster accept connections.
        master_username(str, Optional):
            The name of the master user for the cluster. Constraints: Must be from 1 to 63 letters or numbers.
            The first character must be a letter. Cannot be a reserved word for the chosen database engine.
        master_user_password(str, Optional):
            The password for the master database user. This password can contain any printable ASCII character except
            forward slash (/), double quote ("), or the "at" symbol (@). Constraints: Must contain from 8 to 100 characters.
        preferred_backup_window(str, Optional):
            The daily time range during which automated backups are created if automated backups are enabled
            using the BackupRetentionPeriod parameter. The default is a 30-minute window selected at random
            from an 8-hour block of time for each Amazon Region. To see the time blocks available, see
            Adjusting the Preferred Maintenance Window in the Amazon Doc User Guide.  Constraints:
            Must be in the format hh24:mi-hh24:mi. Must be in Universal Coordinated Time (UTC).   Must not
            conflict with the preferred maintenance window. Must be at least 30 minutes. Defaults to None.
        preferred_maintenance_window(str, Optional):
            The weekly time range during which system maintenance can occur, in Universal Coordinated Time
            (UTC). Format: ddd:hh24:mi-ddd:hh24:mi  The default is a 30-minute window selected at random
            from an 8-hour block of time for each Amazon Region, occurring on a random day of the week. To
            see the time blocks available, see  Adjusting the Preferred Maintenance Window in the Amazon
            Doc User Guide.  Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun. Constraints: Minimum
            30-minute window. Defaults to None.
        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the DB cluster.Each tag consists of a key name and
            an associated value. Defaults to None.

            * Key(str, Optional):
                The key of the tag.
                Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.

            * Value(str, Optional):
                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        storage_encrypted(bool, Optional):
            Specifies whether the DB cluster is encrypted. Defaults to None.
        kms_key_id(str, Optional):
            The KMS key identifier for an encrypted cluster.

            The KMS key identifier is the Amazon Resource Name (ARN) for the KMS encryption key. If you are
            creating a cluster using the same Amazon Web Services account that owns the KMS encryption
            key that is used to encrypt the new cluster, you can use the KMS key alias instead
            of the ARN for the KMS encryption key. If an encryption key is not
            specified in KmsKeyId: If ReplicationSourceIdentifier identifies an encrypted source, then
            Amazon Doc will use the encryption key used to encrypt the source. Otherwise, Amazon Doc
            will use your default encryption key. If the StorageEncrypted parameter is true, then Amazon Doc will use
            your default encryption key. KMS creates the default encryption key for your Amazon Web Services account.
            Your Amazon Web Services account has a different default encryption key for each Amazon Web Services Regions.
        enable_cloudwatch_logs_exports(list, Optional):
            The list of log types that need to be enabled for exporting to CloudWatch Logs. Defaults to None.
        deletion_protection(bool, Optional):
            A value that indicates whether the DB cluster has deletion protection enabled. The database
            can't be deleted when deletion protection is enabled. By default, deletion protection is
            enabled. Defaults to None.
        source_region(str, Optional):
            The ID of the region that contains the source for the db cluster. Defaults to None.
        apply_immediately(bool, Optional):
            A value that indicates whether the modifications in this request and any
            pending modifications are asynchronously applied as soon as possible,
            regardless of the PreferredMaintenanceWindow setting for the DB cluster.
            If this parameter is disabled, changes to the DB cluster are applied during
            the next maintenance window.
            The ApplyImmediately parameter only affects the
            EnableIAMDatabaseAuthentication,MasterUserPassword and NewDBClusterIdentifier
            values.If the ApplyImmediately parameter is disabled, then changes to the
            EnableIAMDatabaseAuthentication,MasterUserPassword and NewDBClusterIdentifier
            values are applied during the next maintenance window. All other changes are
            applied immediately,regardless of the value of the ApplyImmediately parameter
        cloudwatch_logs_export_configuration(dict[str, Any], Optional):
            The configuration setting for the log types to be enabled for export to CloudWatch Logs for a specific DB cluster.

            * EnableLogTypes (list[str], Optional):
                The list of log types to enable.

            * DisableLogTypes(list[str], Optional):
                The list of log types to disable.
        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_doc_db_cluster]:
                aws.docdb.db_cluster.present:
                    - name: 'string'
                    - resource_id: 'string'
                    - engine: 'string'
                    - availability_zones:
                        - 'string'
                        - 'string'
                    - backup_retention_period: 'integer'
                    - db_cluster_parameter_group_name: 'string'
                    - vpc_security_group_ids:
                        - 'string'
                        - 'string'
                    - db_subnet_group_name: 'string'
                    - engine_version: 'string'
                    - port: 'integer'
                    - master_username: 'string'
                    - master_user_password: 'string'
                    - preferred_backup_window: 'string'
                    - preferred_maintenance_window: 'string'
                    - tags:
                        - Key: 'string'
                          Value: 'string'
                        - Key: 'string'
                          Value: 'string'
                    - storage_encrypted: 'boolean'
                    - kms_key_id: 'string'
                    - enable_cloudwatch_logs_exports:
                        - string
                        - string
                    - deletion_protection: 'boolean'
                    - source_region: 'string'
                    - apply_immediately: 'boolean'
                    - cloudwatch_logs_export_configuration:
                        - EnableLogTypes:
                            - 'string'
                            - 'string'
                        - DisableLogTypes:
                            - 'string'
                            - 'string'
                    - timeout:
                        create:
                            delay: 'integer'
                            max_attempts: 'integer'
                        update:
                            delay: 'integer'
                            max_attempts: 'integer'

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            database-1:
              aws.docdb.db_cluster.present:
              - name: database-1
              - resource_id: database-1
              - db_cluster_identifier: database-1
              - engine: docdb
              - availability_zones:
                - ca-central-1a
                - ca-central-1b
                - ca-central-1d
              - backup_retention_period: 1
              - db_cluster_parameter_group_name: default.docdb1
              - db_subnet_group_name: default
              - engine_version: 1.1.1.0
              - port: 8182
              - preferred_backup_window: 06:18-06:48
              - preferred_maintenance_window: wed:04:33-wed:05:03
              - tags:
                  test: test
              - storage_encrypted: true
              - kms_key_id: arn:aws:kms:ca-central-1:746014882121:key/2a4a587a-04fb-4c3b-b4d5-c3f25fb7f69f
              - deletion_protection: false
              - vpc_security_group_ids:
                - sg-ad5a2ac5
    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    desired_state = {
        "name": name,
        "availability_zones": availability_zones,
        "backup_retention_period": backup_retention_period,
        "db_cluster_parameter_group_name": db_cluster_parameter_group_name,
        "vpc_security_group_ids": vpc_security_group_ids,
        "db_subnet_group_name": db_subnet_group_name,
        "engine": engine,
        "engine_version": engine_version,
        "port": port,
        "master_username": master_username,
        "master_user_password": master_user_password,
        "preferred_backup_window": preferred_backup_window,
        "preferred_maintenance_window": preferred_maintenance_window,
        "tags": tags,
        "storage_encrypted": storage_encrypted,
        "kms_key_id": kms_key_id,
        "enable_cloudwatch_logs_exports": enable_cloudwatch_logs_exports,
        "deletion_protection": deletion_protection,
        "source_region": source_region,
    }
    hub.log.debug(f"desired state is {desired_state}")
    is_created = False

    if not resource_id:
        # if there is no resource_id, create db cluster
        # if test run, generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["name"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.docdb.db_cluster", name=name
            )
        else:
            # if this is not a test run, create the real cluster
            create_db_cluster_result = (
                await hub.exec.boto3.client.docdb.create_db_cluster(
                    ctx,
                    DBClusterIdentifier=name,
                    AvailabilityZones=availability_zones,
                    BackupRetentionPeriod=backup_retention_period,
                    DBClusterParameterGroupName=db_cluster_parameter_group_name,
                    VpcSecurityGroupIds=vpc_security_group_ids,
                    DBSubnetGroupName=db_subnet_group_name,
                    Engine=engine,
                    EngineVersion=engine_version,
                    Port=port,
                    MasterUsername=master_username,
                    MasterUserPassword=master_user_password,
                    PreferredBackupWindow=preferred_backup_window,
                    PreferredMaintenanceWindow=preferred_maintenance_window,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                    StorageEncrypted=storage_encrypted,
                    KmsKeyId=kms_key_id,
                    EnableCloudwatchLogsExports=enable_cloudwatch_logs_exports,
                    DeletionProtection=deletion_protection,
                    SourceRegion=source_region,
                )
            )
            # if create failed, fail the request
            if not create_db_cluster_result["result"]:
                result["comment"].append(create_db_cluster_result["comment"])
                result["result"] = False
                return result

            # using the resource_id, wait till db cluster is created
            resource_id = create_db_cluster_result["ret"]["DBCluster"][
                "DBClusterIdentifier"
            ]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="ClusterCreated",
                operation="DescribeDBClusters",
                argument=["DBClusters[].Status"],
                acceptors=create_waiter_acceptors,
                client=await hub.tool.boto3.client.get_client(ctx, "docdb"),
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "docdb",
                    "ClusterCreated",
                    cluster_waiter,
                    DBClusterIdentifier=resource_id,
                    WaiterConfig=waiter_config,
                )
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.docdb.db_cluster", name=name
                )
                result["new_state"] = create_db_cluster_result
            except Exception as e:
                result["comment"] += (str(e),)
                result["result"] = False
                return result
        is_created = True

    # docdb db_cluster has some parameters that can only be applied while modification,
    # so if user specifies those parameters as during creation of a new db cluster, we have to modify
    # the new db cluster to achieve desired state
    desired_state["apply_immediately"] = apply_immediately
    desired_state[
        "cloudwatch_logs_export_configuration"
    ] = cloudwatch_logs_export_configuration

    update_db_cluster_result = await hub.tool.aws.docdb.db_cluster.update_db_cluster(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        desired_state=desired_state,
        is_created=is_created,
        update_waiter_acceptors=update_waiter_acceptors,
        timeout=timeout,
    )
    if not (
        update_db_cluster_result
        and update_db_cluster_result["result"]
        and update_db_cluster_result["new_state"]
    ):
        result["result"] = False
        result["comment"].append(update_db_cluster_result["comment"])
        return result

    result["old_state"] = update_db_cluster_result["old_state"]
    result["new_state"] = update_db_cluster_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] = update_db_cluster_result["comment"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    skip_final_snapshot: bool = None,
    final_db_snapshot_identifier: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """The DeleteDBCluster action deletes a previously provisioned DB cluster.

    When you delete a DB cluster, all automated backups for that DB cluster are deleted and can't be recovered.
    Manual DB cluster snapshots of the specified DB cluster are not deleted.
    Note that the DB Cluster cannot be deleted if deletion protection is
    enabled. To delete it, you must first set its DeletionProtection field to False.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            An identifier of the resource in the provider. The DB cluster identifier for the DB cluster to be deleted. This parameter isn't case-sensitive.
            Idem automatically considers this resource being absent if this field is not specified.
            Constraints:   Must match an existing DBClusterIdentifier.
        skip_final_snapshot(bool, Optional):
            Determines whether a final DB cluster snapshot is created before the DB cluster is deleted. If
            true is specified, no DB cluster snapshot is created. If false is specified, a DB cluster
            snapshot is created before the DB cluster is deleted.  You must specify a
            FinalDBSnapshotIdentifier parameter if SkipFinalSnapshot is false.  Default: false. Defaults to None.
        final_db_snapshot_identifier(str, Optional):
            The DB cluster snapshot identifier of the new DB cluster snapshot created when
            SkipFinalSnapshot is set to false.   Specifying this parameter and also setting the
            SkipFinalShapshot parameter to true results in an error.  Constraints:   Must be 1 to 255
            letters, numbers, or hyphens.   First character must be a letter   Cannot end with a hyphen or
            contain two consecutive hyphens. Defaults to None.
        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_doc_db_cluster]:
                aws.docdb.db_cluster.absent:
                    - name: 'string'
                    - resource_id: 'string'
                    - skip_final_snapshot: 'boolean'
                    - final_db_snapshot_identifier: 'string'
                    - timeout:
                        delete:
                            delay: int
                            max_attempts: int

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws_auto.docdb.db_cluster.absent:
                - name: database-1
                - resource_id: database-1
                - skip_final_snapshot: True
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.docdb.db_cluster", name=name
    )

    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.docdb.db_cluster.get(ctx=ctx, name=resource_id)

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        return result

    result["old_state"] = before["ret"]

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.docdb.db_cluster", name=name
        )
        return result

    ret = await hub.exec.boto3.client.docdb.delete_db_cluster(
        ctx,
        DBClusterIdentifier=resource_id,
        SkipFinalSnapshot=skip_final_snapshot,
        FinalDBSnapshotIdentifier=final_db_snapshot_identifier,
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]
        return result

    # Custom waiter for delete
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get("delete") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="ClusterDelete",
        operation="DescribeDBClusters",
        argument=["Error.Code", "DBClusters[].Status"],
        acceptors=delete_waiter_acceptors,
        client=await hub.tool.boto3.client.get_client(ctx, "docdb"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "docdb",
            "ClusterDelete",
            cluster_waiter,
            DBClusterIdentifier=resource_id,
            WaiterConfig=waiter_config,
        )
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.docdb.db_cluster", name=name
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns information about provisioned Doc DB clusters.  This operation can also return
    information for Amazon RDS clusters and Amazon DocDB clusters.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.docdb.db_cluster
    """
    result = {}
    ret = await hub.exec.aws.docdb.db_cluster.list(ctx=ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.docdb.db_cluster {ret['comment']}")
        return {}

    for db_cluster in ret["ret"]:
        resource_id = db_cluster.get("db_cluster_identifier")
        result[resource_id] = {"aws.docdb.db_cluster.present": [db_cluster]}

    return result



from typing import Any
from typing import Dict


async def get(
    hub,
    ctx,
    *,
    name: str,
    route_table_id: str = None,
    destination_cidr_block: str = None,
    destination_ipv6_cidr_block: str = None,
    destination_prefix_list_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Check if a route is attached to a route table.
    If resource_id is received as input search is done using only resource_id, If other parameters are specified
    they will be ignored. If resource_id is not specified then route_table_id and one of the destination should be
    specified. If more than one destination is specified search is done using only one destination in the priority order
    destination_cidr_block, destination_ipv6_cidr_block, destination_prefix_list_id

    Args:
        name(str): An Idem name of the resource
        route_table_id(str, Optional): The ID of the route table for the route.
        destination_cidr_block(str, Optional): The IPv4 CIDR address block used for the destination match. Routing decisions are based on the
            most specific match. We modify the specified CIDR block to its canonical form; for example, if
            you specify 100.68.0.18/18, we modify it to 100.68.0.0/18. Defaults to None.
        destination_ipv6_cidr_block(str, Optional): The IPv6 CIDR block used for the destination match. Routing decisions are based on the most
            specific match. Defaults to None.
        destination_prefix_list_id(str, Optional): The ID of a prefix list used for the destination match. Defaults to None.
        resource_id(str, Optional): Unique identifier to identify the route in a route table. format of resource id is
            <route_table_id>/<destination_cidr_block or destination_ipv6_cidr_block or destination_prefix_list_id>

    Returns
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The EC2 route in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ec2.route.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.ec2.route.get(
                    ctx, name=name, resource_id=resource_id
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.route.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id

    """
    result = dict(comment=[], result=True, ret=None)
    if resource_id:
        route_table_id, route_destination = resource_id.split("/", 1)
    elif route_table_id and (
        destination_cidr_block
        or destination_ipv6_cidr_block
        or destination_prefix_list_id
    ):
        if destination_cidr_block:
            route_destination = destination_cidr_block
        elif destination_ipv6_cidr_block:
            route_destination = destination_ipv6_cidr_block
        else:
            route_destination = destination_prefix_list_id
    else:
        result["result"] = False
        result["comment"] = [
            f"aws.ec2.route {name} either resource_id or both route table id and one of the destination"
            f" should be specified."
        ]
        return result

    ret = await hub.tool.aws.ec2.route_table.search_raw(
        ctx=ctx, resource_id=route_table_id
    )
    if not ret["result"]:
        if "InvalidRouteTableID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.route", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["RouteTables"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.route", name=name
            )
        )
        return result

    resource = ret["ret"]["RouteTables"][0]
    routes_associated_with_route_table = resource["Routes"]
    route_found = None
    for route in routes_associated_with_route_table:
        if route["State"] == "active":
            if (
                "DestinationCidrBlock" in route
                and route_destination == route["DestinationCidrBlock"]
            ):
                route_found = route
                break
            elif (
                "DestinationIpv6CidrBlock" in route
                and route_destination == route["DestinationIpv6CidrBlock"]
            ):
                route_found = route
                break
            elif (
                "DestinationPrefixListId" in route
                and route_destination == route["DestinationPrefixListId"]
            ):
                route_found = route
                break

    if route_found:
        result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_to_present(
            raw_resource=route_found,
            route_table_id=route_table_id,
            idem_resource_name=name,
        )
    else:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.route", name=name
            )
        )
    return result



"""Exec module for managing Amazon ECR Images."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    repository_name: str,
    registry_id: str = None,
    image_digest: str = None,
    image_tag: str = None,
) -> Dict[str, Any]:
    """Retrieves the specified AWS ECR Image details.

    Args:
        name(str):
            The name of the Idem state.
        repository_name(str):
            The name of repository that contains the image.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository that contains the images.
            If you do not specify a registry, the default registry is assumed.
        image_digest(str, Optional):
            The sha256 digest of the image manifest. Either `image_digest` or `image_tag` has to be provided.
        image_tag(str, Optional):
            The tag used for the image. Either `image_digest` or `image_tag` has to be provided.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.image.get name='idem_name' repository_name='ecr_repository_name' image_tag='ecr_image_tag'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.image.get
                - kwargs:
                    name: 'idem_name'
                    repository_name: 'ecr_repository_name'
                    image_tag: 'ecr_image_tag'
    """
    result = dict(comment=[], ret=None, result=True)

    image_id = {}
    if image_digest:
        image_id["imageDigest"] = image_digest
    elif image_tag:
        image_id["imageTag"] = image_tag
    else:
        result["result"] = False
        result["comment"].append(
            "Either `image_digest` or `image_tag` has to be provided"
        )
        return result

    ret = await hub.exec.boto3.client.ecr.describe_images(
        ctx, registryId=registry_id, repositoryName=repository_name, imageIds=[image_id]
    )
    if not ret["result"]:
        if "RepositoryNotFoundException" in str(
            ret["comment"]
        ) or "ImageNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ecr.image", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("imageDetails"):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ecr.image", name=name
            )
        )
        return result

    if len(ret["ret"]["imageDetails"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ecr.image", resource_id=name
            )
        )

    result["ret"] = hub.tool.aws.ecr.conversion_utils.convert_raw_image_to_present(
        raw_resource=ret["ret"]["imageDetails"][0],
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    repository_name: str,
    registry_id: str = None,
) -> Dict[str, Any]:
    """Lists AWS ECR Images for a Repository.

    Arg:
        repository_name(str):
            The name of repository with images to be listed.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository in which to list images.
            If you do not specify a registry, the default registry is assumed.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.image.list repository_name='ecr_repository_name'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.image.list
                - kwargs:
                    repository_name: 'ecr_repository_name'
    """
    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.ecr.describe_images(
        ctx=ctx,
        repositoryName=repository_name,
        registryId=registry_id,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("imageDetails"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ecr.image", name=repository_name
            )
        )
        return result

    for image in ret["ret"]["imageDetails"]:
        result["ret"].append(
            hub.tool.aws.ecr.conversion_utils.convert_raw_image_to_present(
                raw_resource=image
            )
        )

    return result



"""State module to manage AWS SES Identity notification topic."""
import copy
import re
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    identity: str,
    notification_type: str,
    sns_topic: str = None,
    enabled: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Sets an Amazon Simple Notification Service (Amazon SNS) topic to use when delivering notifications.

    When you use this operation, you specify a verified identity, such as an email address or domain. When you send an
    email that uses the chosen identity in the Source field, Amazon SES sends notifications to the topic you specified.

    Args:
        name(str):
            A name for the Identity Notification Topic.

        identity(str):
            The identity (email address or domain) that you want to set the Amazon SNS topic for.

        notification_type(str):
            The type of notifications that will be published to the specified Amazon SNS topic.

        sns_topic(str, Optional):
            The Amazon Resource Name (ARN) of the Amazon SNS topic. If the parameter is omitted from the request or a
            null value is passed, SnsTopic is cleared and publishing is disabled.

        enabled(bool, Optional):
            Sets whether Amazon SES includes the original email headers in Amazon SNS notifications of the specified
            notification type. A value of true specifies that Amazon SES will include headers in notifications, and a
            value of false specifies that Amazon SES will not include headers in notifications.

        resource_id(str, Optional):
            The identifier for AWS Identity Notification Topic, combination of identity, notification topic separated
            by a separator '::'.

    Request Syntax:
        .. code-block:: sls

            resource_is_present:
              aws.ses.identity_notification_topic.present:
                - name: "string"
                - identity: "string"
                - notification_type: "string"
                - sns_topic: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.ses.identity_notification_topic.present:
                - name: name
                - identity: user@example.com
                - notification_type: Bounce
                - sns_topic: arn:aws:sns:us-west-2:111122223333:MyTopic
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    changed = False

    if resource_id:
        if (
            not re.findall("::", resource_id)
            or not len(re.findall("::", resource_id)) == 1
        ):
            result[
                "comment"
            ] = f"Incorrect aws.ses.identity_notification_topic resource_id: {resource_id}. Expected id <identity>::<notification_type>"
            result["result"] = False
            return result

        before = await hub.exec.aws.ses.identity_notification_topic.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

    if before and before.get("ret"):
        result["old_state"] = before["ret"].copy()
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.ses.identity_notification_topic", name=name
        )
        if (
            result["old_state"]["sns_topic"] != sns_topic
            or result["old_state"].get("enabled") != enabled
        ):
            changed = True
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

    if sns_topic or enabled:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "identity": identity,
                    "notification_type": notification_type,
                    "sns_topic": sns_topic,
                    "enabled": enabled,
                },
            )
            if changed:
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.ses.identity_notification_topic", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.ses.identity_notification_topic", name=name
                )
            return result

        if sns_topic:
            ret = await hub.exec.boto3.client.ses.set_identity_notification_topic(
                ctx,
                Identity=identity,
                NotificationType=notification_type,
                SnsTopic=sns_topic,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] += ret["comment"]
                return result
            if changed:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ses.identity_notification_topic", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.create_comment(
                    "aws.ses.identity_notification_topic", name
                )
        if enabled is not None:
            ret = await hub.exec.boto3.client.ses.set_identity_headers_in_notifications_enabled(
                ctx,
                Identity=identity,
                NotificationType=notification_type,
                Enabled=enabled,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] += ret["comment"]
                return result
            if not (before and before["ret"]) or changed:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ses.identity_notification_topic", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.create_comment(
                    "aws.ses.identity_notification_topic", name
                )
        new = await hub.exec.aws.ses.identity_notification_topic.get(
            ctx,
            name=name,
            resource_id=resource_id if changed else identity + "::" + notification_type,
        )
        if not new["result"]:
            result["comment"] += new["comment"]
            result["result"] = False
            return result

        result["new_state"] = new["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Clears the SnsTopic and disables the publishing.

    Args:
        name(str):
            A name for the Identity Notification Topic.

        resource_id(str, Optional):
            The identifier for AWS Identity Notification Topic, combination of identity, notification topic separated by
            a separator '::'.

    Request Syntax:
        .. code-block:: sls

            resource_is_absent:
              aws.ses.identity_notification_topic.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ses.identity_notification_topic.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ses.identity_notification_topic", name=name
        )
        return result
    identity = resource_id.split("::")[0]
    notification_type = resource_id.split("::")[1]
    before = await hub.exec.aws.ses.identity_notification_topic.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if before.get("ret"):
        result["old_state"] = before["ret"].copy()
        if ctx.get("test", False):
            result["comment"] += getattr(
                hub.tool.aws.comment_utils, "would_delete_comment"
            )(resource_type="aws.ses.identity_notification_topic", name=name)
            return result
        else:
            ret = await hub.exec.boto3.client.ses.set_identity_notification_topic(
                ctx,
                Identity=identity,
                NotificationType=notification_type,
                SnsTopic=None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] += ret["comment"]
                result["result"] = False
                return result
            result["comment"] += getattr(hub.tool.aws.comment_utils, "delete_comment")(
                resource_type="aws.ses.identity_notification_topic", name=name
            )

    else:
        result["comment"] += getattr(
            hub.tool.aws.comment_utils, "already_absent_comment"
        )(resource_type="aws.ses.identity_notification_topic", name=name)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            idem describe aws.ses.identity_notification_topic
    """

    result = {}

    ret = await hub.exec.aws.ses.identity_notification_topic.list(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not list identities {ret['comment']}")
        return {}

    for resource in ret["ret"]:

        resource_id = resource.get("resource_id")

        result[resource_id] = {
            "aws.ses.identity_notification_topic.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result
